import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.channels.SocketChannel;
import java.nio.channels.Selector;
import java.nio.channels.SelectionKey;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CodingErrorAction;
import java.io.IOException;

class Connection
{
    static final Charset charset = Charset.forName ("US-ASCII");
    static final CharsetDecoder charsetDecoder = charset.newDecoder ().onMalformedInput (CodingErrorAction.IGNORE);
    static final CharsetEncoder charsetEncoder = charset.newEncoder ().onMalformedInput (CodingErrorAction.IGNORE);
          
    ByteBuffer readBuffer, writeBuffer;
    Selector selector;
    SelectionKey selectionKey;
    SocketChannel channel;

    Connection ()
    {
        readBuffer = ByteBuffer.allocateDirect (4096);
        writeBuffer = ByteBuffer.allocateDirect (4096);
        try
        {
            selector = Selector.open ();
        }
        catch (IOException e)
        {
            System.out.println (e.toString ());
        }
    }

    boolean isDisconnected ()
    {
        return channel == null;
    }

    boolean isConnected ()
    {
        return channel != null && channel.isConnected ();
    }

    boolean isConnecting ()
    {
        return channel != null && channel.isConnectionPending ();
    }

    synchronized String disconnect ()
    {
        if (isDisconnected ())
            return "\034F9\035Not connected.\034C\035\n";
        boolean connecting = isConnecting ();
        try
        {
            channel.close ();
        }
        catch (IOException e)
        {
        }
        channel = null;
        writeBuffer.clear ();
        return connecting ? "\034F15\035Connection attempt aborted.\034C\035\n" : "\034F15\035Disconnected.\034C\035\n";
    }

    synchronized String connect (String host, int port)
    {
        if (isConnecting ())
            return "\034F9\035Already attempting connection.\034C\035\n";
        if (isConnected ())
            return "\034F9\035Already connected.\034C\035\n";
        selector.wakeup ();
        try
        {
            InetSocketAddress address = new InetSocketAddress (InetAddress.getByName (host), port);
            channel = SocketChannel.open ();
            channel.configureBlocking (false);
            channel.connect (address);
            selectionKey = channel.register (selector, SelectionKey.OP_CONNECT);
        }
        catch (IOException e)
        {
            return e.getMessage () + ".\n";
        }
        return "\034F11\035Connecting to " + host + " " + (new Integer (port)).toString () + "...\034C\035\n";
    }
    
    String finishConnect ()
    {
        try
        {
            if (channel.finishConnect ())
            {
                selectionKey.interestOps (SelectionKey.OP_READ);
                return "\034F10\035Connected.\034C\035\n";
            }
            return null;
        }
        catch (IOException e)
        {
            disconnect ();
            return "\034F9\035" + e.getMessage () + ".\034C\035\n";
        }
    }
    
    String read ()
    {
        int bytesRead;
        try
        {
            writeBuffer.flip ();
            channel.write (writeBuffer);
            writeBuffer.compact ();
            readBuffer.clear ();
            bytesRead = channel.read (readBuffer);
        }
        catch (IOException e)
        {
            return "\034F9\035" + e.getMessage () + ".\034C\035\n" + disconnect ();
        }
        if (bytesRead < 0)
            return disconnect ();
        if (bytesRead == 0)
            return null;
        try
        {
            readBuffer.flip ();
            return charsetDecoder.decode (readBuffer).toString ();
        }
        catch (Exception e)
        {
            return null;
        }
    }        

    String poll ()
    {
        try
        {
            if (writeBuffer.position () == 0)
                selector.select ();
        }   
        catch (IOException e)
        {
            return "\034F9\035" + e.getMessage () + ".\034C\035\n" + disconnect ();
        }
        synchronized (this)
        {
            if (isConnecting ())
                return finishConnect ();
            if (isConnected ())
                return read ();
        }
        return null;
    }
                
    synchronized String write (String text)
    {

        if (isConnecting ())
            return "\034F9\035Attempting connection.\034C\035\n";
        if (! isConnected ())
            return "\034F9\035Not connected.\034C\035\n";
        try
        {
            charsetEncoder.encode (CharBuffer.wrap (text), writeBuffer, false);
        }
        catch (Exception e)
        {
        }
        selector.wakeup ();
        return text;
    }
}
    
