import javax.swing.JComponent;
import javax.swing.JFrame;
import java.awt.Point;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Image;
import java.awt.Font;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseEvent;
import java.util.TreeMap;

class Terminal extends JComponent
    implements ClipboardOwner, MouseListener, MouseMotionListener
{
    static final int BORDER = 2;
    static final int MIN_ROWS = 1;//24;
    static final int MIN_COLS = 1;//80;
    static final int TAB_COLS = 4;
 
    int columns, rows;
    int scrollLimit, scrollOffset, currentLine, visibleLine;
    LinkedList [] lines;
    int cellWidth, cellHeight, cellOffset;
    Font font; 
    int foregroundColor, backgroundColor;
    boolean ansiBold;
    Color [] colors;
    Action action;
    Point selectStart, selectEnd;
    UI ui;
    TreeMap commands;
    Connection connection;
    StringBuffer commandBuffer;
    TreeMap aliases;

    Terminal (UI u, int sl)
    {
        scrollLimit = sl;
        scrollOffset = 0;
        currentLine = 0;
        visibleLine = 0;
        lines = new LinkedList [scrollLimit];
        for (int line = 0; line < scrollLimit; ++ line)
            lines [line] = new LinkedList ();
        foregroundColor = 7;
        backgroundColor = 0;
        ansiBold = false;
        colors = new Color []
        { 
            new Color (0, 0, 0),
            new Color (170, 0, 0),
            new Color (0, 170, 0),
            new Color (170, 85, 0),
            new Color (0, 0, 170),
            new Color (170, 0, 170), 
            new Color (0, 170, 170),
            new Color (170, 170, 170),
            new Color (85, 85, 85),
            new Color (255, 85, 85),
            new Color (85, 255, 85),
            new Color (255, 255, 85),
            new Color (85, 85, 255),
            new Color (255, 85, 255),
            new Color (85, 255, 255),
            new Color (255, 255, 255)
        };
        ui = u;
        commands = new TreeMap ();
        commands.put ("help", new HelpCommand ());
        commands.put ("font", new FontCommand ());
        commands.put ("alias", new AliasCommand ());
        commands.put ("unalias", new UnaliasCommand ());
        commands.put ("echo", new EchoCommand ());
        commands.put ("echon", new EchoNCommand ());
        commands.put ("copy", new CopyCommand ());
        commands.put ("paste", new PasteCommand ());
        commands.put ("connect", new ConnectCommand ());
        commands.put ("disconnect", new DisconnectCommand ());
        commands.put ("quit", new QuitCommand ());
        connection = new Connection ();
        commandBuffer = new StringBuffer ();
        aliases = new TreeMap ();
            
        setFont ("Courier", 12);

        addMouseListener (this);
        addMouseMotionListener (this);

        setFocusable (false);
        setVisible (true);
    }

    void test ()
    {
        for (int i = 0; i < colors.length; ++ i) 
        { 
            foregroundColor = i; 
            addText ("abcdefghijklmnopqrstuvwxyz".substring (i, i + 1)); 
        }
        foregroundColor = 5;
        action = new Hyperlink ("/echon fooey");
        addText ("foo\n");
        foregroundColor = 4;
        action = new Hyperlink ("/echo \034Iimage.jpg;1;1\035");
        addText ("bar\n");
        foregroundColor = 3;
        action = new Hyperlink ("/echo \034F3\035\n/help");
        addText ("help\n");
        foregroundColor = 2;
        action = new Hyperlink ("/help echo");
        addText ("echo\n");
        foregroundColor = 1;
        action = new Hyperlink ("/echon Imagey images imagine imagery imaging imaginations!\n/echo \034Iimage.jpg;10;6\035");
        addImage ("image.jpg", 5, 3);
        foregroundColor = 7;
    }

    void run ()
    {
        while (true) 
        { 
            String text = connection.poll (); 
            if (text != null) 
                addText (text);
        }
    }

    void updateCells ()
    {
        FontMetrics fm = getFontMetrics (font);

        cellHeight = fm.getMaxDescent () + fm.getMaxAscent () + fm.getLeading () + 1;
        cellWidth = fm.charWidth ('#');
        cellOffset = fm.getMaxDescent ();

        int terminalWidth = getWidth (),
            terminalHeight = getHeight ();
        columns = Math.max ((getWidth () - 2 * BORDER) / cellWidth, MIN_COLS);
        rows = Math.max ((getHeight () - 2 * BORDER) / cellHeight, MIN_ROWS);

        updateScrollBar ();

        repaint ();
    }

    void setFont (String name, int size)
    {
        font = new Font (name, Font.PLAIN, size);
        updateCells ();
    }

    void updateScrollBar ()
    {
        ui.updateScrollBar (rows, visibleLine + 1);
    }

    void setScrollOffset (int offset)
    {
        scrollOffset = Math.max (Math.min (offset, visibleLine - rows + 1), 0);
        repaint ();
    }

    private void scrollLines (int count)
    {
        currentLine -= count;
        visibleLine -= count;
        System.arraycopy (lines, count, lines, 0, scrollLimit - count);
        while (count > 0)
        {
            lines [scrollLimit - count].clear ();
            -- count;
        }
        updateScrollBar ();
    }
        
    private void newLine ()
    {
        ++ currentLine;
        visibleLine = Math.max (visibleLine, currentLine);
        if (currentLine >= scrollLimit)
            scrollLines (1);
        else
            updateScrollBar ();
    }

    private int addFragment (int startColumn, int row, Fragment fragment)
    {
        if (row < 0 || row >= scrollLimit)
            return -1;
        LinkedList line = lines [row];
        int column = 0;
        for (Link link = line.first (); link != line.end (); link = link.nextLink)
        {
            Fragment nextFragment = (Fragment) link;
            if (startColumn < nextFragment.column)
            {
                column = Math.max (startColumn, column);
                if (nextFragment.column - column >= fragment.width ())
                {
                    fragment.column = column;
                    fragment.insertBefore (nextFragment);
                    visibleLine = Math.max (visibleLine, row);
                    return column;
                }
            }
            column = nextFragment.endColumn ();
        }
        column = Math.max (startColumn, column);
        if (columns - column >= fragment.width ())
        {
            fragment.column = column;
            line.addLast (fragment);
            visibleLine = Math.max (visibleLine, row);
            return column;
        } 
        return -1;
    }

    synchronized void addImage (String name, int width, int height)
    {
        if (height > scrollLimit - currentLine)
            scrollLines (height - (scrollLimit - currentLine));
        Image image = ui.cacheImage (name);
        ImageFragment fragment = new ImageFragment (action, 0, width, height, 0, image);
        int column = addFragment (0, currentLine, fragment);
        while (column < 0)
        {
            newLine ();
            column = addFragment (0, currentLine, fragment);
        }
        for (int offset = 1; offset < height; ++ offset)
            addFragment (column, currentLine + offset, new ImageFragment (action, column, width, height, offset, image));
        repaint ();
    }    
        
    int parseAnsi (String text, int index)
    {
        char c = ' ';
        while (index < text.length ())
        {
            c = text.charAt (index ++);
            if (Character.isLetter (c))
                break;
            commandBuffer.append (c);
        }
        if (! Character.isLetter (c))
            return index;
        switch (c)
        {
        case 'm':
            {
                int pos = 2;
                for (;;)
                {
                    int sep = commandBuffer.indexOf(";", pos);
                    int val = Integer.parseInt (commandBuffer.substring (pos, sep < 0 ? commandBuffer.length () : sep));
                    switch (val)
                    {
                    case 0:
                        foregroundColor = 7;
                        backgroundColor = 0;
                        ansiBold = false;
                        break;
                    case 1: ansiBold = true; break;
                    case 2: ansiBold = false; break;

                    case 30: foregroundColor = 0 + (ansiBold ? 8 : 0); break;
                    case 31: foregroundColor = 1 + (ansiBold ? 8 : 0); break;
                    case 32: foregroundColor = 2 + (ansiBold ? 8 : 0); break;
                    case 33: foregroundColor = 3 + (ansiBold ? 8 : 0); break;
                    case 34: foregroundColor = 4 + (ansiBold ? 8 : 0); break;
                    case 35: foregroundColor = 5 + (ansiBold ? 8 : 0); break;
                    case 36: foregroundColor = 6 + (ansiBold ? 8 : 0); break;
                    case 37: foregroundColor = 7 + (ansiBold ? 8 : 0); break;

                    case 40: backgroundColor = 0; break;
                    case 41: backgroundColor = 1; break;
                    case 42: backgroundColor = 2; break;
                    case 43: backgroundColor = 3; break;
                    case 44: backgroundColor = 4; break;
                    case 45: backgroundColor = 5; break;
                    case 46: backgroundColor = 6; break;
                    case 47: backgroundColor = 7; break;
                    }
                    
                    if(sep < 0) break;
                    pos = sep + 1;
                }
            }
        }
        commandBuffer.delete (0, commandBuffer.length ());
        return index;
    }

    int parseCommand (String text, int index)
    {
        char c = ' ';
        while (index < text.length ())
        {
            c = text.charAt (index ++);
            if (c == '\035')
                break;
            commandBuffer.append (c);
        }
        if (c != '\035')
            return index;
        char command = 'C';
        if (commandBuffer.length () >= 2)
          command = commandBuffer.charAt (1);
        switch (command)
        {
        case 'C':
            foregroundColor = 7;
            backgroundColor = 0;
            action = null;
            break;
        
        case 'B':
            {
                int sep = commandBuffer.indexOf (";", 2);
                if (sep >= 0)
                    ui.addButton (commandBuffer.substring (2, sep), commandBuffer.substring (sep + 1));
                else
                    ui.removeButton (commandBuffer.substring (2));
            }
            break;
                    
        case 'H':
            {
                String arg = commandBuffer.substring (2);
                if (arg.length () == 0)
                    action = null;
                else
                    action = new Hyperlink (arg);
            }
            break;

        case 'F':
            {
                String arg = commandBuffer.substring (2);
                if (arg.length () == 0)
                    foregroundColor = 7;
                else
                    foregroundColor = Integer.parseInt (arg);
            }
            break;

        case 'G':
            {
                String arg = commandBuffer.substring (2);
                if (arg.length () == 0)
                    backgroundColor = 0;
                else
                    backgroundColor = Integer.parseInt (arg);
            }
            break;
 
        case 'I':
            {
                int sep1 = commandBuffer.indexOf (";", 2), 
                    sep2 = commandBuffer.indexOf (";", sep1 + 1);
                if (sep1 < 0 || sep2 < 0)
                    break;
                String arg1 = commandBuffer.substring (2, sep1);
                int arg2 = Math.min (columns, Math.max (1, Integer.parseInt (commandBuffer.substring (sep1 + 1, sep2)))),
                    arg3 = Math.min (rows, Math.max (1, Integer.parseInt (commandBuffer.substring (sep2 + 1)))); 
                addImage (arg1, arg2, arg3);
            }
            break;
        }
        commandBuffer.delete (0, commandBuffer.length ());
        return index;
    }
             
    void addText (String text)
    {
        addText (text, true);
    }

    synchronized void addText (String text, boolean parseCommands)
    {
        int index = 0;
        if(parseCommands && commandBuffer.length () > 0)
        {
            switch(commandBuffer.charAt (0))
            {
            case '\034': index = parseCommand (text, 0); break;
            case '\033': index = parseAnsi (text, 0); break;
            }
        }

        StringBuffer buffer = new StringBuffer ();
        while (index < text.length () || buffer.length () > 0)
        {
            if (parseCommands)
            {
                while (index < text.length () && (text.charAt (index) == '\034' || text.charAt (index) == '\033'))
                {
                    commandBuffer.append (text.charAt (index));
                    switch(text.charAt (index))
                    {
                        case '\034': index = parseCommand (text, index + 1); break;
                        case '\033': index = parseAnsi (text, index + 1); break;
                    }
                }
                if (index >= text.length () && buffer.length () == 0)
                    break;
            }

            LinkedList line = lines [currentLine];
            int column = 0;
            Fragment pos = null;

            for (Link link = line.first (); link != line.end (); link = link.nextLink)
            {
                Fragment fragment = (Fragment) link;
                if (column < fragment.column)
                {
                    pos = fragment;
                    break;
                }
                column = fragment.endColumn ();
            }

            int space = (pos == null ? columns : pos.column) - column;
            endLine:
            while (index < text.length () && buffer.length () < space)
            {
                char c = text.charAt (index ++);
                switch (c)
                {
                case '\034':
                case '\033':
                    if (parseCommands)
                    {
                        -- index;
                        break endLine;
                    }
                    else
                        break;

                case '\n':
                    newLine ();
                    break endLine;

                case '\t':
                    do buffer.append (' ');
                    while (((column + buffer.length ()) % TAB_COLS) > 0);
                    break;

                default:
                    if (c >= ' ' && c <= '~')
                        buffer.append (c); 
                    break;
                }
            }

            int length = Math.min (buffer.length (), space);
            if (length == 0)
            {
                if (space == 0)
                    newLine ();
                continue;
            }
            TextFragment textFragment = new TextFragment (action, column, colors [foregroundColor], colors [backgroundColor], buffer.substring (0, length));
            buffer.setLength (buffer.length () - length);
            if (pos == null)
            {
                line.addLast (textFragment);
                if (space == length)
                  newLine ();
            }
            else
            if (! pos.merge (textFragment))
            {
                textFragment.insertBefore (pos);
            }
        }

        repaint ();
    }

    boolean input (String text)
    {
        if (text.length () == 0)
            return false;    
        int index = 0;
        while (index < text.length ())
        {
            int end = text.indexOf ('\n', index);
            if (end < 0)
                end = text.length ();
            String line = text.substring (index, end);
            index = end + 1;
            if (line.charAt (0) == '/')
            {
                String [] args = Command.parseCommand (line);
                String name = args [0];
                if (name.equals("test"))
                {
                    test ();
                }
                else if (commands.containsKey (name))
                {
                    Command command = (Command) commands.get (name);
                    if (! command.execute (this, args))
                        addText ("\"\034F9\035" + text + "\034C\035\" is invalid usage. See \"\034H/help " + name + "\035\034F6\035/help " + name + "\034C\035\" for usage information.\n");
                }
                else
                if (aliases.containsKey (name))
                {
                    String expansion = AliasCommand.expandAlias ((String) aliases.get (name), args);
                    return input (expansion);
                }
                else
                    addText ("There is no command \"\034F9\035" + name + "\034C\035\". See \"\034H/help\035\034F6\035/help\034C\035\" for available commands.\n");
            }
            else
                addText (connection.write (line + "\n"));
        }
        return true;
    }

    int topLine ()
    {
        return Math.max (visibleLine - scrollOffset - rows + 1, 0);
    }

    int bottomLine ()
    {
        return topLine () + rows;
    }

    public void paint (Graphics g)
    {
        g.setColor (Color.black);
        g.fillRect (0, 0, getWidth (), getHeight ());

        g.setFont (font);

        Graphics xor = g.create();
        xor.setColor(Color.cyan);
        xor.setXORMode(Color.magenta);

        int sx1 = -1, sy1 = -1, sx2 = -1, sy2 = -1;

        if (selectStart != selectEnd)
        {
            sx1 = Math.min (selectStart.x, selectEnd.x);
            sy1 = Math.min (selectStart.y, selectEnd.y);
            sx2 = Math.max (selectStart.x, selectEnd.x);
            sy2 = Math.max (selectStart.y, selectEnd.y);
        }

        for (int topLine = topLine (), offset = 0; offset < rows; ++ offset)
        {
            LinkedList line = lines [topLine + offset];
            int column = 0, cursor = -1;
            for (Link link = line.first (); link != line.end (); link = link.nextLink)
            {
                Fragment fragment = (Fragment) link;
                fragment.draw (g, this, offset);
                if (topLine + offset == currentLine && column < fragment.column && cursor < 0)
                    cursor = column;
                if (fragment.maySelect () && topLine + offset >= sy1 && topLine + offset <= sy2)
                {
                    int x1 = (topLine + offset == sy1 ? Math.max (sx1, fragment.column) : fragment.column), 
                        x2 = (topLine + offset == sy2 ? Math.min (sx2 + 1, fragment.endColumn ()) : fragment.endColumn ());
                    if (x1 < x2)
                    {
                        xor.fillRect (x1 * cellWidth + BORDER, offset * cellHeight + BORDER,
                                      (x2 - x1) * cellWidth, cellHeight);
                    }
                }
                column = fragment.endColumn ();
            }
            if (topLine + offset == currentLine && column < columns && cursor < 0)
                cursor = column;
            if (cursor >= 0)
            {
                g.setColor (colors [foregroundColor]);
                g.fillRect (cursor * cellWidth + BORDER, offset * cellHeight + BORDER, cellWidth, cellHeight);
            }
        }                

        xor.dispose ();
    }
   
    public String getSelectedText ()
    {
        if (selectStart == selectEnd)
            return null;
        StringBuffer buffer = new StringBuffer ();
        int sx1 = Math.min (selectStart.x, selectEnd.x),
            sy1 = Math.min (selectStart.y, selectEnd.y),
            sx2 = Math.max (selectStart.x, selectEnd.x),
            sy2 = Math.max (selectStart.y, selectEnd.y);
        for (int row = sy1, maxRow = sy2 + 1; row < maxRow; ++ row)
        {
            if (row != sy1)
                buffer.append ('\n');
            LinkedList line = lines [row];
            for (Link link = line.first (); link != line.end (); link = link.nextLink)
            {
                Fragment fragment = (Fragment) link;
               
                if (fragment.maySelect ())
                {
                    int x1 = (row == sy1 ? Math.max (sx1, fragment.column) : fragment.column),
                        x2 = (row == sy2 ? Math.min (sx2 + 1, fragment.endColumn ()) : fragment.endColumn ());
                    if (x1 < x2)
                        fragment.select (buffer, x1 - fragment.column, x2 - x1);
                }
            }
        }
        
        return buffer.toString ();
    }
 
    private Point pointToCell (Point p)
    {
        return new Point ((p.x - BORDER) / cellWidth, topLine () + (p.y - BORDER) / cellHeight);
    }

    private Fragment fragmentContaining (int column, int row)
    {
        if (row < 0 || row >= scrollLimit)
            return null;
        LinkedList line = lines [row];
        for (Link link = line.first (); link != line.end (); link = link.nextLink)
        {
            Fragment fragment = (Fragment) link;
            if (column < fragment.column)
                return null;
            if (column < fragment.endColumn ())
                return fragment;
        }
        return null;
    }

// ClipboardOwner
    public void lostOwnership (Clipboard clipboard, Transferable contents) 
    { 
    }

// MouseListener
    public void mouseClicked (MouseEvent event)
    {
    }

    public void mouseEntered (MouseEvent event)
    {
    }

    public void mouseExited (MouseEvent event)
    {
    }

    public void mousePressed (MouseEvent event)
    {
        Point point = event.getPoint ();
        if (! contains (point))
            return;

        selectStart = pointToCell (point);
        selectEnd = selectStart; 

        repaint ();
    }

    public void mouseReleased (MouseEvent event)
    {
        Point point = event.getPoint ();
        if (! contains (point))
            return;
 
        if (selectStart == selectEnd)
        {
            int column = (point.x - BORDER) / cellWidth,
                row = topLine () + (point.y - BORDER) / cellHeight;
            Fragment fragment = fragmentContaining (column, row);
            if (fragment != null)
                fragment.activate (this);
        }
        else
        {
            selectEnd = pointToCell (point);
            Clipboard selectionBuffer = getToolkit ().getSystemSelection ();
            if (selectionBuffer != null)
                selectionBuffer.setContents (new StringSelection (getSelectedText ()), this);
        }
        repaint ();
    }

// MouseMotionListener
    public void mouseDragged (MouseEvent event)
    {
        Point point = event.getPoint ();
        if (! contains (point))
            return;

        selectEnd = pointToCell (point);
        repaint ();
    }

    public void mouseMoved (MouseEvent event)
    {
    }
}
